@@ -2,10 +2,10 @@
 -- anchored at a certain root (this is usually the Tree root). They are
 -- represented by a list of Names (these are just strict bytestrings).
 module Storage.Hashed.AnchoredPath
-    ( Name(..), AnchoredPath(..), appendPath, anchorPath
+    ( Name(..), AnchoredPath(..), anchoredRoot, appendPath, anchorPath
     , isPrefix, parent, parents, catPaths, flatten, makeName
     -- * Unsafe functions.
-    , nameToFilePath, nameFromFilePath, floatBS, floatPath ) where
+    , floatBS, floatPath ) where
 
 import qualified Data.ByteString.Char8 as BS
 import Data.List( isPrefixOf, inits )
@@ -16,16 +16,15 @@ import System.FilePath( (</>), splitDirectories, normalise, dropTrailingPathSepa
 --
 
 newtype Name = Name BS.ByteString  deriving (Eq, Show, Ord)
+
+-- | This is a type of "sane" file paths. These are always canonic in the sense
+-- that there are no stray slashes, no ".." components and similar. They are
+-- usually used to refer to a location within a Tree, but a relative filesystem
+-- path works just as well. These are either constructed from individual name
+-- components (using "appendPath", "catPaths" and "makeName"), or converted
+-- from a FilePath ("floatPath" -- but take care when doing that) or .
 newtype AnchoredPath = AnchoredPath [Name] deriving (Eq, Show, Ord)
 
--- | Unsafe.
-nameToFilePath :: Name -> FilePath
-nameToFilePath (Name p) = BS.unpack p
-
--- | Unsafe.
-nameFromFilePath :: FilePath -> Name
-nameFromFilePath = Name . BS.pack
-
 -- | Check whether a path is a prefix of another path.
 isPrefix :: AnchoredPath -> AnchoredPath -> Bool
 (AnchoredPath a) `isPrefix` (AnchoredPath b) = a `isPrefixOf` b
@@ -35,6 +34,7 @@ appendPath :: AnchoredPath -> Name -> AnchoredPath
 appendPath (AnchoredPath p) n =
     case n of
       (Name s) | s == BS.empty -> AnchoredPath p
+               | s == BS.pack "." -> AnchoredPath p
                | otherwise -> AnchoredPath $ p ++ [n]
 
 -- | Catenate two paths together. Not very safe, but sometimes useful
@@ -58,6 +58,8 @@ anchorPath :: FilePath -> AnchoredPath -> FilePath
 anchorPath dir p = dir </> BS.unpack (flatten p)
 {-# INLINE anchorPath #-}
 
+-- | Unsafe. Only ever use on bytestrings that came from flatten on a
+-- pre-existing AnchoredPath.
 floatBS :: BS.ByteString -> AnchoredPath
 floatBS = AnchoredPath . map Name . takeWhile (not . BS.null) . BS.split '/'
 
@@ -67,12 +69,22 @@ flatten (AnchoredPath p) = BS.intercalate (BS.singleton '/')
                                            [ n | (Name n) <- p ]
 
 makeName :: String -> Name
-makeName = Name . BS.pack
+makeName ".." = error ".. is not a valid AnchoredPath component name"
+makeName n | '/' `elem` n = error "/ may not occur in a valid AnchoredPath component name"
+           | otherwise = Name $ BS.pack n
 
 -- | Take a relative FilePath and turn it into an AnchoredPath. The operation
--- is unsafe and if you break it, you keep both pieces. More useful for
--- exploratory purposes (ghci) than for serious programming.
+-- is (relatively) unsafe. Basically, by using floatPath, you are testifying
+-- that the argument is a path relative to some common root -- i.e. the root of
+-- the associated "Tree" object. Also, there are certain invariants about
+-- AnchoredPath that this function tries hard to preserve, but probably cannot
+-- guarantee (i.e. this is a best-effort thing). You should sanitize any
+-- FilePaths before you declare them "good" by converting into AnchoredPath
+-- (using this function).
 floatPath :: FilePath -> AnchoredPath
 floatPath = AnchoredPath . map (Name . BS.pack) . splitDirectories
             . normalise . dropTrailingPathSeparator
 
+
+anchoredRoot :: AnchoredPath
+anchoredRoot = AnchoredPath []
