@@ -1,18 +1,25 @@
 {-# LANGUAGE ScopedTypeVariables, FlexibleInstances #-}
 module Storage.Hashed.Test( tests ) where
 
-import Prelude hiding ( read, filter )
+import Prelude hiding ( filter, readFile, writeFile )
+import qualified Prelude
 import qualified Data.ByteString.Lazy.Char8 as BL
 import qualified Data.ByteString.Char8 as BS
 import Control.Exception( finally )
 import System.Process
-import System.Directory( doesFileExist, removeFile )
+import System.Directory( doesFileExist, removeFile, doesDirectoryExist )
+import System.FilePath( (</>) )
 import Control.Monad( forM_, when )
 import Control.Monad.Identity
+import Control.Monad.Trans( lift )
+import Control.Applicative( (<$>) )
+
 import Data.Maybe
 import Data.Word
+import Data.Int
 import Data.Bits
-import Data.List( (\\), sort, intercalate, nub )
+import Data.List( (\\), sort, intercalate, nub, intersperse )
+
 import Storage.Hashed
 import Storage.Hashed.AnchoredPath
 import Storage.Hashed.Tree
@@ -20,6 +27,9 @@ import Storage.Hashed.Index
 import Storage.Hashed.Utils
 import Storage.Hashed.Darcs
 import Storage.Hashed.Packed
+import Storage.Hashed.Hash
+import Storage.Hashed.Monad hiding ( tree )
+
 import System.IO.Unsafe( unsafePerformIO )
 import System.Mem( performGC )
 
@@ -54,23 +64,27 @@ dirs = [ floatPath "foo_dir"
        , floatPath "foo_dir/foo_subdir"
        , floatPath "foo space" ]
 
-emptyStub = Stub (return emptyTree) Nothing
+emptyStub = Stub (return emptyTree) NoHash
 
 testTree =
     makeTree [ (makeName "foo", emptyStub)
              , (makeName "subtree", SubTree sub)
-             , (makeName "substub", Stub getsub Nothing) ]
+             , (makeName "substub", Stub getsub NoHash) ]
     where sub = makeTree [ (makeName "stub", emptyStub)
-                         , (makeName "substub", Stub getsub2 Nothing)
+                         , (makeName "substub", Stub getsub2 NoHash)
                          , (makeName "x", SubTree emptyTree) ]
           getsub = return sub
-          getsub2 = return $ makeTree [ (makeName "file", File emptyBlob) ]
+          getsub2 = return $ makeTree [ (makeName "file", File emptyBlob)
+                                      , (makeName "file2",
+                                         File $ Blob (return $ BL.pack "foo") NoHash) ]
 
 equals_testdata t = sequence_ [
-                     do ours <- read b
-                        let Just stored = Prelude.lookup p blobs
-                        assertEqual "contents match" ours stored
-                     | (p, File b) <- list t ]
+                     do isJust (findFile t p) @? show p ++ " in tree"
+                        ours <- readBlob (fromJust $ findFile t p)
+                        ours @?= stored
+                     | (p, stored) <- blobs ] >>
+                    sequence_ [ isJust (Prelude.lookup p blobs) @? show p ++ " extra in tree"
+                                | (p, File _) <- list t ]
 
 ---------------------------
 -- Test list
@@ -78,9 +92,11 @@ equals_testdata t = sequence_ [
 
 tests = [ testGroup "Bundled.Posix" posix
         , testGroup "Storage.Hashed.Utils" utils
+        , testGroup "Storage.Hashed.Hash" hash
         , testGroup "Storage.Hashed.Tree" tree
         , testGroup "Storage.Hashed.Index" index
         , testGroup "Storage.Hashed.Packed" packed
+        , testGroup "Storage.Hashed.Monad" monad
         , testGroup "Storage.Hashed" hashed ]
 
 --------------------------
@@ -95,7 +111,7 @@ hashed = [ testCase "plain has all files" have_files
          , testCase "writePlainTree works" write_plain ]
     where
       check_file t f = assertBool
-                       ("path " ++ show f ++ " is missing in tree")
+                       ("path " ++ show f ++ " is missing in tree " ++ show t)
                        (isJust $ find t f)
       check_files = forM_ files . check_file
 
@@ -118,26 +134,26 @@ hashed = [ testCase "plain has all files" have_files
       write_plain = do
         orig <- readDarcsPristine "." >>= expand
         writePlainTree orig "_darcs/plain"
-        t <- readPlainTree "_darcs/plain"
+        t <- expand =<< readPlainTree "_darcs/plain"
         equals_testdata t
 
-index = [ testCase "index listing" check_index
-        , testCase "index content" check_index_content
-        , testCase "index versioning" check_index_versions ]
+index = [ testCase "index versioning" check_index_versions
+        , testCase "index listing" check_index
+        , testCase "index content" check_index_content ]
     where pristine = readDarcsPristine "." >>= expand
           build_index =
             do x <- pristine
                exist <- doesFileExist "_darcs/index"
                performGC -- required in win32 to trigger file close
                when exist $ removeFile "_darcs/index"
-               idx <- updateIndexFrom "_darcs/index" darcsTreeHash x >>= expand
+               idx <- updateIndex =<< updateIndexFrom "_darcs/index" darcsTreeHash x
                return (x, idx)
           check_index =
             do (pris, idx) <- build_index
                (sort $ map fst $ list idx) @?= (sort $ map fst $ list pris)
           check_blob_pair p x y =
-              do a <- read x
-                 b <- read y
+              do a <- readBlob x
+                 b <- readBlob y
                  assertEqual ("content match on " ++ show p) a b
           check_index_content =
             do (_, idx) <- build_index
@@ -146,11 +162,9 @@ index = [ testCase "index listing" check_index
                assertBool "files match" (length x > 0)
           check_index_versions =
             do performGC -- required in win32 to trigger file close
-               writeFile "_darcs/index" "nonsense index... do not crash!"
-               pris <- pristine
-               idx <- expand =<<
-                        readOrUpgradeIndex "_darcs/index" darcsTreeHash pristine
-               (sort $ map fst $ list idx) @?= (sort $ map fst $ list pris)
+               Prelude.writeFile "_darcs/index" "nonsense index... do not crash!"
+               valid <- indexFormatValid "_darcs/index"
+               assertBool "index format invalid" $ not valid
 
 tree = [ testCase "modifyTree" check_modify
        , testCase "complex modifyTree" check_modify_complex
@@ -159,19 +173,22 @@ tree = [ testCase "modifyTree" check_modify
        , testCase "expandPath" check_expand_path
        , testCase "diffTrees" check_diffTrees
        , testCase "diffTrees identical" check_diffTrees_ident
-       , testProperty "treeEq" prop_tree_eq
+       , testProperty "shapeEq" prop_shape_eq
+       , testProperty "expandedShapeEq" prop_expanded_shape_eq
        , testProperty "expand is identity" prop_expand_id
        , testProperty "filter True is identity" prop_filter_id
        , testProperty "filter False is empty" prop_filter_empty
        , testProperty "restrict both ways keeps shape" prop_restrict_shape_commutative
-       , testProperty "restrict is a subtree of both" prop_restrict_subtree ]
-    where blob x = File $ Blob (return (BL.pack x)) (Just $ sha256 $ BL.pack x)
+       , testProperty "restrict is a subtree of both" prop_restrict_subtree
+       , testProperty "overlay keeps shape" prop_overlay_shape
+       , testProperty "overlay is superset of over" prop_overlay_super ]
+    where blob x = File $ Blob (return (BL.pack x)) (sha256 $ BL.pack x)
           name = Name . BS.pack
           check_modify =
               let t = makeTree [(name "foo", blob "bar")]
                   modify = modifyTree t (floatPath "foo") (Just $ blob "bla")
-               in do x <- read $ fromJust $ findFile t (floatPath "foo")
-                     y <- read $ fromJust $ findFile modify (floatPath "foo")
+               in do x <- readBlob $ fromJust $ findFile t (floatPath "foo")
+                     y <- readBlob $ fromJust $ findFile modify (floatPath "foo")
                      assertEqual "old version" x (BL.pack "bar")
                      assertEqual "new version" y (BL.pack "bla")
                      assertBool "list has foo" $
@@ -182,11 +199,11 @@ tree = [ testCase "modifyTree" check_modify
                                , (name "bar", SubTree t1) ]
                   t1 = makeTree [ (name "foo", blob "bar") ]
                   modify = modifyTree t (floatPath "bar/foo") (Just $ blob "bla")
-               in do foo <- read $ fromJust $ findFile t (floatPath "foo")
-                     foo' <- read $ fromJust $ findFile modify (floatPath "foo")
-                     bar_foo <- read $ fromJust $
+               in do foo <- readBlob $ fromJust $ findFile t (floatPath "foo")
+                     foo' <- readBlob $ fromJust $ findFile modify (floatPath "foo")
+                     bar_foo <- readBlob $ fromJust $
                                 findFile t (floatPath "bar/foo")
-                     bar_foo' <- read $ fromJust $
+                     bar_foo' <- readBlob $ fromJust $
                                  findFile modify (floatPath "bar/foo")
                      assertEqual "old foo" foo (BL.pack "bar")
                      assertEqual "old bar/foo" bar_foo (BL.pack "bar")
@@ -199,8 +216,8 @@ tree = [ testCase "modifyTree" check_modify
                      length (list modify) @?= length (list t)
           check_modify_remove =
               let t1 = makeTree [(name "foo", blob "bar")]
-                  t2 = makeTree [ (name "foo", blob "bar")
-                                , (name "bar", SubTree t1) ]
+                  t2 :: Tree Identity = makeTree [ (name "foo", blob "bar")
+                                                 , (name "bar", SubTree t1) ]
                   modify1 = modifyTree t1 (floatPath "foo") Nothing
                   modify2 = modifyTree t2 (floatPath "bar") Nothing
                   file = findFile modify1 (floatPath "foo")
@@ -237,19 +254,21 @@ tree = [ testCase "modifyTree" check_modify
                        badpath `notElem` (map fst $ list t')
 
           check_diffTrees =
-            flip finally (writeFile "foo_dir/foo_a" "a\n") $
-                 do writeFile "foo_dir/foo_a" "b\n"
+            flip finally (Prelude.writeFile "foo_dir/foo_a" "a\n") $
+                 do Prelude.writeFile "foo_dir/foo_a" "b\n"
                     working_plain <- filter nondarcs `fmap` readPlainTree "."
-                    working <- expand =<< updateIndexFrom "_darcs/index" darcsTreeHash working_plain
+                    working <- updateIndex =<<
+                                 updateIndexFrom "_darcs/index" darcsTreeHash working_plain
                     pristine <- readDarcsPristine "."
                     (working', pristine') <- diffTrees working pristine
                     let foo_work = findFile working' (floatPath "foo_dir/foo_a")
                         foo_pris = findFile pristine' (floatPath "foo_dir/foo_a")
-                    assertBool "trees have equal shapes" (working' `treeEq` pristine')
+                    working' `shapeEq` pristine'
+                             @? show working' ++ " `shapeEq` " ++ show pristine'
                     assertBool "foo_dir/foo_a is in working'" $ isJust foo_work
                     assertBool "foo_dir/foo_a is in pristine'" $ isJust foo_pris
-                    foo_work_c <- read (fromJust foo_work)
-                    foo_pris_c <- read (fromJust foo_pris)
+                    foo_work_c <- readBlob (fromJust foo_work)
+                    foo_pris_c <- readBlob (fromJust foo_pris)
                     BL.unpack foo_work_c @?= "b\n"
                     BL.unpack foo_pris_c @?= "a\n"
                     assertEqual "working' tree is minimal" 2 (length $ list working')
@@ -261,43 +280,62 @@ tree = [ testCase "modifyTree" check_modify
             assertBool "t1 is empty" $ null (list t1)
             assertBool "t2 is empty" $ null (list t2)
 
-          prop_tree_eq x = x `treeEq` x
-          prop_expand_id x = unsafePerformIO (expand x) `treeEq` x
-          prop_filter_id x = filter (\_ _ -> True) x `treeEq` x
-          prop_filter_empty x = filter (\_ _ -> False) x `treeEq` emptyTree
+          prop_shape_eq x = no_stubs x ==> x `shapeEq` x
+              where types = x :: Tree Identity
+          prop_expanded_shape_eq x = runIdentity $ expandedShapeEq x x
+              where types = x :: Tree Identity
+          prop_expand_id x = no_stubs x ==> runIdentity (expand x) `shapeEq` x
+              where types = x :: Tree Identity
+          prop_filter_id x = runIdentity $ expandedShapeEq x $ filter (\_ _ -> True) x
+              where types = x :: Tree Identity
+          prop_filter_empty x = runIdentity $ expandedShapeEq emptyTree $ filter (\_ _ -> False) x
+              where types = x :: Tree Identity
           prop_restrict_shape_commutative (t1, t2) =
-              not (restrict t1 t2 `treeEq` emptyTree) ==>
-                  restrict t1 t2 `treeEq` restrict t2 t1
+              no_stubs t1 && no_stubs t2 && not (restrict t1 t2 `shapeEq` emptyTree) ==>
+                  restrict t1 t2 `shapeEq` restrict t2 t1
+              where types = (t1 :: Tree Identity, t2 :: Tree Identity)
           prop_restrict_subtree (t1, t2) =
-              not (restrict t1 t2 `treeEq` emptyTree) ==>
+              no_stubs t1 && not (restrict t1 t2 `shapeEq` emptyTree) ==>
                   let restricted = S.fromList (map fst $ list $ restrict t1 t2)
                       orig1 = S.fromList (map fst $ list t1)
                       orig2 = S.fromList (map fst $ list t2)
                    in and [restricted `S.isSubsetOf` orig1, restricted `S.isSubsetOf` orig2]
+              where types = (t1 :: Tree Identity, t2 :: Tree Identity)
+          prop_overlay_shape (t1 :: Tree Identity, t2) =
+              (Just LT == runIdentity (t2 `cmpExpandedShape` t1)) ==>
+              runIdentity $ (t1 `overlay` t2) `expandedShapeEq` t1
+          prop_overlay_super (t1 :: Tree Identity, t2) =
+              (Just LT == runIdentity (t2 `cmpExpandedShape` t1)) ==>
+              Just EQ == (runIdentity $ restrict t2 (t1 `overlay` t2) `cmpTree` t2)
+
 
 packed = [ testCase "loose pristine tree" check_loose
-         , testCase "readOS" check_readOS
+         , testCase "load" check_load
          , testCase "live" check_live
          , testCase "compact" check_compact ]
-    where check_loose = do x <- readDarcsPristine "."
+    where root_hash = treeHash <$> get_pristine
+          get_pristine = darcsUpdateDirHashes <$> (expand =<< readDarcsPristine ".")
+          check_loose = do x <- readDarcsPristine "."
                            os <- create "_darcs/loose" Loose
                            (os', root) <- writePackedDarcsPristine x os
-                           y <- readPackedDarcsPristine os' root
+                           y <- expand =<< readPackedDarcsPristine os' root
                            equals_testdata y
-          check_readOS = do os <- readOS "_darcs/loose"
-                            format (hatchery os) @?= Loose
-                            x <- readDarcsPristine "."
-                            y <- readPackedDarcsPristine os (fromJust $ treeHash x)
-                            equals_testdata y
-          check_live = do os <- readOS "_darcs/loose"
-                          x <- readDarcsPristine "."
-                          alive <- live (os { roots = [ fromJust $ treeHash x ]
-                                            , references = darcsPristineRefs } ) [hatchery os]
-                          sequence_ [ assertBool "" $ (fromJust hash) `S.member` M.keysSet alive
+          check_load = do os <- load "_darcs/loose"
+                          format (hatchery os) @?= Loose
+                          root <- root_hash
+                          y <- expand =<< readPackedDarcsPristine os root
+                          equals_testdata y
+          check_live = do os <- load "_darcs/loose"
+                          x <- get_pristine
+                          root <- root_hash
+                          alive <- live (os { roots = [ root ]
+                                            , references = darcsPristineRefs }) [hatchery os]
+                          sequence_ [ assertBool (show hash ++ " is alive") $
+                                                 hash `S.member` M.keysSet alive
                                       | hash <- map (itemHash . snd) $ list x ]
                           length (M.toList alive) @?= 1 + length (nub $ map snd blobs) + length dirs
-          check_compact = do os <- readOS "_darcs/loose"
-                             x <- darcsUpdateHashes `fmap` (expand =<< readDarcsPristine ".")
+          check_compact = do os <- load "_darcs/loose"
+                             x <- darcsUpdateDirHashes `fmap` (expand =<< readDarcsPristine ".")
                              (os', root) <- storePackedDarcsPristine x os
                              hatch_root_old <- blockLookup (hatchery os') root
                              assertBool "bits in the old hatchery" $ isJust hatch_root_old
@@ -305,14 +343,19 @@ packed = [ testCase "loose pristine tree" check_loose
                              os'' <- compact os'
                              length (mature os'') @?= 1
                              hatch_root <- blockLookup (hatchery os'') root
+                             mature_root <- blockLookup (head $ mature os'') root
                              assertBool "bits no longer in hatchery" $ isNothing hatch_root
+                             assertBool "bits now in the mature space" $ isJust mature_root
+                             mature_root_con <- readSegment (fromJust mature_root)
+                             Just mature_root_con @?= darcsFormatDir x
 
-                             -- TODO:
-                             -- y <- readPackedDarcsPristine os'' (fromJust $ treeHash x)
-                             -- equals_testdata y
+                             y <- expand =<< readPackedDarcsPristine os'' root
+                             equals_testdata y
 
 utils = [ testProperty "xlate32" prop_xlate32
         , testProperty "xlate64" prop_xlate64
+        , testProperty "align bounded" prop_align_bounded
+        , testProperty "align aligned" prop_align_aligned
         , testProperty "reachable is a subset" prop_reach_subset
         , testProperty "roots are reachable" prop_reach_roots
         , testProperty "nonexistent roots are not reachable" prop_reach_nonroots
@@ -321,6 +364,14 @@ utils = [ testProperty "xlate32" prop_xlate32
         , testCase "mmap empty file" check_mmapEmpty ]
     where prop_xlate32 x = (xlate32 . xlate32) x == x where types = x :: Word32
           prop_xlate64 x = (xlate64 . xlate64) x == x where types = x :: Word64
+          prop_align_bounded (bound, x) =
+              bound > 0 && bound < 1024 && x >= 0 ==>
+                    align bound x >= x && align bound x < x + bound
+              where types = (bound, x) :: (Int, Int)
+          prop_align_aligned (bound, x) =
+              bound > 0 && bound < 1024 && x >= 0 ==>
+                    align bound x `rem` bound == 0
+              where types = (bound, x) :: (Int, Int)
 
           check_fixFrom = let f 0 = 0
                               f n = f (n - 1) in fixFrom f 5 @?= 0
@@ -360,6 +411,18 @@ utils = [ testProperty "xlate32" prop_xlate32
                                   Nothing -> Nothing
                                   Just i -> Just $ M.elemAt i m
 
+hash = [ testProperty "decodeBase16 . encodeBase16 == id" prop_base16
+       , testProperty "decodeBase64u . encodeBase64u == id" prop_base64u ]
+    where prop_base16 x = (decodeBase16 . encodeBase16) x == x
+          prop_base64u x = (decodeBase64u . encodeBase64u) x == x
+
+monad = [ testCase "path expansion" check_virtual ]
+    where check_virtual = virtualTreeMonad run testTree >> return ()
+              where run = do file <- readFile (floatPath "substub/substub/file")
+                             file2 <- readFile (floatPath "substub/substub/file2")
+                             lift $ BL.unpack file @?= ""
+                             lift $ BL.unpack file2 @?= "foo"
+
 posix = [ testCase "getFileStatus" $ check_stat Posix.getFileStatus
         , testCase "getSymbolicLinkStatus" $ check_stat Posix.getSymbolicLinkStatus ]
     where check_stat fun = flip finally (removeFile "test_empty") $ do
@@ -383,10 +446,26 @@ instance (Arbitrary a, Ord a) => Arbitrary (S.Set a)
 instance (Arbitrary k, Arbitrary v, Ord k) => Arbitrary (M.Map k v)
     where arbitrary = M.fromList `fmap` arbitrary
 
+instance Arbitrary BL.ByteString where
+    arbitrary = BL.pack `fmap` arbitrary
+
 instance Arbitrary Word32 where
     arbitrary = do x <- arbitrary :: Gen Int
                    return $ fromIntegral x
 
+instance Arbitrary Word8 where
+    arbitrary = do x <- arbitrary :: Gen Int
+                   return $ fromIntegral x
+
+instance Arbitrary Hash where
+    arbitrary = sized hash'
+        where hash' 0 = return NoHash
+              hash' _ = do
+                tag <- oneof [return 0, return 1]
+                case tag of
+                  0 -> SHA256 . BS.pack <$> sequence [ arbitrary | _ <- [1..32] ]
+                  1 -> SHA1 . BS.pack <$> sequence [ arbitrary | _ <- [1..20] ]
+
 instance Arbitrary Word64 where
     arbitrary = do x <- arbitrary :: Gen Int
                    y <- arbitrary :: Gen Int
@@ -394,26 +473,48 @@ instance Arbitrary Word64 where
                        y' = fromIntegral y
                    return $ x' .|. (y' `shift` 32)
 
-instance Arbitrary TreeItem where
+instance Arbitrary Int64 where
+    arbitrary = fromIntegral `fmap` (arbitrary :: Gen Word64)
+
+instance (Monad m) => Arbitrary (TreeItem m) where
   arbitrary = sized tree'
     where tree' 0 = oneof [ return (File emptyBlob), return (SubTree emptyTree) ]
-          tree' n = do branches <- choose (1, n)
-                       let subtree name = do t <- tree' ((n - 1) `div` branches)
-                                             return (makeName $ show name, t)
-                       sublist <- mapM subtree [0..branches]
-                       oneof [ tree' 0
-                             , return (SubTree $ makeTree sublist) ]
+          tree' n = oneof [ file n, subtree n ]
+          file 0 = return (File emptyBlob)
+          file _ = do content <- arbitrary
+                      return (File $ Blob (return content) NoHash)
+          subtree n = do branches <- choose (1, n)
+                         let sub name = do t <- tree' ((n - 1) `div` branches)
+                                           return (makeName $ show name, t)
+                         sublist <- mapM sub [0..branches]
+                         oneof [ tree' 0
+                               , return (SubTree $ makeTree sublist)
+                               , return $ (Stub $ return (makeTree sublist)) NoHash ]
 
-instance Arbitrary Tree where
+instance (Monad m) => Arbitrary (Tree m) where
   arbitrary = do item <- arbitrary
                  case item of
                    File _ -> arbitrary
+                   Stub _ _ -> arbitrary
                    SubTree t -> return t
 
 ---------------------------
 -- Other instances
 --
 
+instance Show (Blob m) where
+    show (Blob _ h) = "Blob " ++ show h
+
+instance Show (TreeItem m) where
+    show (File f) = "File (" ++ show f ++ ")"
+    show (Stub _ h) = "Stub _ " ++ show h
+    show (SubTree s) = "SubTree (" ++ show s ++ ")"
+
+instance Show (Tree m) where
+    show t = "Tree " ++ show (treeHash t) ++ " { " ++
+             (concat . intersperse ", " $ itemstrs) ++ " }"
+        where itemstrs = map show $ listImmediate t
+
 instance Show (Int -> Int) where
     show f = "[" ++ intercalate ", " (map val [1..20]) ++ " ...]"
         where val x = show x ++ " -> " ++ show (f x)
@@ -422,13 +523,62 @@ instance Show (Int -> Int) where
 -- Test utilities
 --
 
-treeItemEq (File _) (File _) = True
-treeItemEq (SubTree s) (SubTree p) = s `treeEq` p
-treeItemEq _ _ = False
+shapeEq a b = Just EQ == cmpShape a b
+expandedShapeEq a b = (Just EQ ==) <$> cmpExpandedShape a b
 
-treeEq t r = and $ zipTrees cmp t r
-    where cmp _ (Just a) (Just b) = a `treeItemEq` b
-          cmp _ _ _ = False
+cmpcat (x:y:rest) | x == y = cmpcat (x:rest)
+                  | x == Just EQ = cmpcat (y:rest)
+                  | y == Just EQ = cmpcat (x:rest)
+                  | otherwise = Nothing
+cmpcat [x] = x
+cmpcat [] = Just EQ -- empty things are equal
+
+cmpTree a b = do a' <- expand a
+                 b' <- expand b
+                 con <- contentsEq a' b'
+                 return $ cmpcat [cmpShape a' b', con]
+    where contentsEq a b = cmpcat <$> sequence (zipTrees cmp a b)
+          cmp _ (Just (File a)) (Just (File b)) = do a' <- readBlob a
+                                                     b' <- readBlob b
+                                                     return $ Just (compare a' b')
+          cmp _ _ _ = return (Just EQ) -- neutral
+
+cmpShape t r = cmpcat $ zipTrees cmp t r
+    where cmp _ (Just a) (Just b) = a `item` b
+          cmp _ Nothing (Just _) = Just LT
+          cmp _ (Just _) Nothing = Just GT
+          item (File _) (File _) = Just EQ
+          item (SubTree s) (SubTree p) = s `cmpShape` p
+          item _ _ = Nothing
+
+cmpExpandedShape :: (Monad m) => Tree m -> Tree m -> m (Maybe Ordering)
+cmpExpandedShape a b = do x <- expand a
+                          y <- expand b
+                          return $ x `cmpShape` y
 
 nondarcs (AnchoredPath (Name x:_)) _ | x == BS.pack "_darcs" = False
                                      | otherwise = True
+
+readDarcsPristine :: FilePath -> IO (Tree IO)
+readDarcsPristine dir = do
+  let darcs = dir </> "_darcs"
+      h_inventory = darcs </> "hashed_inventory"
+  repo <- doesDirectoryExist darcs
+  unless repo $ fail $ "Not a darcs repository: " ++ dir
+  hashed <- doesFileExist h_inventory
+  if hashed
+     then do inv <- BS.readFile h_inventory
+             let lines = BS.split '\n' inv
+             case lines of
+               [] -> return emptyTree
+               (pris_line:_) -> do
+                          let hash = decodeDarcsHash $ BS.drop 9 pris_line
+                              size = decodeDarcsSize $ BS.drop 9 pris_line
+                          when (hash == NoHash) $ fail $ "Bad pristine root: " ++ show pris_line
+                          readDarcsHashed (darcs </> "pristine.hashed") (size, hash)
+     else do have_pristine <- doesDirectoryExist $ darcs </> "pristine"
+             have_current <- doesDirectoryExist $ darcs </> "current"
+             case (have_pristine, have_current) of
+               (True, _) -> readPlainTree $ darcs </> "pristine"
+               (False, True) -> readPlainTree $ darcs </> "current"
+               (_, _) -> fail "No pristine tree is available!"
