@@ -5,7 +5,6 @@
 module Storage.Hashed.Utils where
 
 import Prelude hiding ( lookup, catch )
-import qualified Bundled.SHA256 as SHA
 import System.Mem( performGC )
 import System.IO.MMap( mmapFileByteString )
 import Bundled.Posix( getFileStatus, fileSize )
@@ -21,47 +20,40 @@ import Control.Applicative( (<$>) )
 import Foreign.ForeignPtr( withForeignPtr )
 import Foreign.Ptr( plusPtr )
 import Data.ByteString.Internal( toForeignPtr, memcpy )
-
+import System.IO (withFile, IOMode(ReadMode), hSeek, SeekMode(AbsoluteSeek))
 import Data.Bits( Bits )
 #ifdef BIGENDIAN
 import Data.Bits( (.&.), (.|.), shift, shiftL, shiftR )
 #endif
 
 import qualified Data.ByteString.Lazy.Char8 as BL
-import qualified Data.ByteString.Char8 as BS
+import qualified Data.ByteString.Char8 as BS8
+import qualified Data.ByteString as BS
 
 import qualified Data.Set as S
 import qualified Data.Map as M
 
-newtype Hash = Hash (Maybe Int64, BS.ByteString) deriving (Show, Eq, Ord, Read)
-
-makeHash :: BS.ByteString -> Hash
-makeHash str = case BS.split '-' str of
-                 [h] -> Hash (Nothing, h)
-                 [s, h] -> Hash (Just $ read $ BS.unpack s, h)
-                 _ -> Hash (Nothing, str) -- ...
-
-hashSetSize :: Hash -> Int64 -> Hash
-hashSetSize (Hash (_,h)) s = Hash (Just s, h)
-
 -- | Pointer to a filesystem, possibly with start/end offsets. Supposed to be
 -- fed to (uncurry mmapFileByteString) or similar.
 type FileSegment = (FilePath, Maybe (Int64, Int))
 
--- | Bad and ugly. Only works well with single-chunk BL's.
-sha256 :: BL.ByteString -> Hash
-sha256 = makeHash . BS.pack . SHA.sha256 . BS.concat . BL.toChunks
-
 -- | Read in a FileSegment into a Lazy ByteString. Implemented using mmap.
 readSegment :: FileSegment -> IO BL.ByteString
 readSegment (f,range) = do
- x <- mmapFileByteString f range
-   `catch` (\(_::SomeException) -> do
+    bs <- tryToRead
+       `catch` (\(_::SomeException) -> do
                      size <- fileSize `fmap` getFileStatus f
                      if size == 0
-                        then return BS.empty
-                        else performGC >> mmapFileByteString f range)
- return $ BL.fromChunks [x]
+                        then return BS8.empty
+                        else performGC >> tryToRead)
+    return $ BL.fromChunks [bs]
+  where
+    tryToRead = do 
+        case range of
+            Nothing -> BS.readFile f
+            Just (off, size) -> withFile f ReadMode $ \h -> do
+                hSeek h AbsoluteSeek $ fromIntegral off
+                BS.hGet h size 
 {-# INLINE readSegment #-}
 
 -- | Run an IO action with @path@ as a working directory. Does neccessary
@@ -87,11 +79,11 @@ makeAbsolute p = do
   return $! isAbsolute p ? (p, cwd </> p)
 
 -- Wow, unsafe.
-pokeBS :: BS.ByteString -> BS.ByteString -> IO ()
-pokeBS to from =
+unsafePokeBS :: BS8.ByteString -> BS8.ByteString -> IO ()
+unsafePokeBS to from =
     do let (fp_to, off_to, len_to) = toForeignPtr to
            (fp_from, off_from, len_from) = toForeignPtr from
-       when (len_to /= len_from) $ fail $ "Length mismatch in pokeBS: from = "
+       when (len_to /= len_from) $ fail $ "Length mismatch in unsafePokeBS: from = "
             ++ show len_from ++ " /= to = " ++ show len_to
        withForeignPtr fp_from $ \p_from ->
          withForeignPtr fp_to $ \p_to ->
@@ -99,6 +91,12 @@ pokeBS to from =
                   (plusPtr p_from off_from)
                   (fromIntegral len_to)
 
+align :: Integral a => a -> a -> a
+align boundary i = case i `rem` boundary of
+                     0 -> i
+                     x -> i + boundary - x
+{-# INLINE align #-}
+
 xlate32 :: (Bits a) => a -> a
 xlate64 :: (Bits a) => a -> a
 
